@@ -40,6 +40,7 @@
 #include "gdkinternals.h"
 #include "gdkdisplay-x11.h"
 #include "gdkkeysyms.h"
+#include "gdkalias.h"
 
 #ifdef HAVE_XKB
 #include <X11/XKBlib.h>
@@ -52,16 +53,26 @@
 #  endif
 #endif /* HAVE_XKB */
 
-typedef struct _GdkKeymapX11 GdkKeymapX11;
+typedef struct _GdkKeymapX11   GdkKeymapX11;
+typedef struct _GdkKeymapClass GdkKeymapX11Class;
 
 #define GDK_TYPE_KEYMAP_X11          (gdk_keymap_x11_get_type ())
 #define GDK_KEYMAP_X11(object)       (G_TYPE_CHECK_INSTANCE_CAST ((object), GDK_TYPE_KEYMAP_X11, GdkKeymapX11))
 #define GDK_IS_KEYMAP_X11(object)    (G_TYPE_CHECK_INSTANCE_TYPE ((object), GDK_TYPE_KEYMAP_X11))
 
+typedef struct _DirectionCacheEntry DirectionCacheEntry;
+
+struct _DirectionCacheEntry
+{
+  guint serial;
+  Atom group_atom;
+  PangoDirection direction;
+};
+
 struct _GdkKeymapX11
 {
   GdkKeymap     parent_instance;
-  
+
   gint min_keycode;
   gint max_keycode;
   KeySym* keymap;
@@ -70,21 +81,35 @@ struct _GdkKeymapX11
   guint lock_keysym;
   GdkModifierType group_switch_mask;
   GdkModifierType num_lock_mask;
+  GdkModifierType modmap[8];
   gboolean sun_keypad;
   PangoDirection current_direction;
   gboolean have_direction;
   guint current_serial;
-  
+
 #ifdef HAVE_XKB
   XkbDescPtr xkb_desc;
+  /* We cache the directions */
+  Atom current_group_atom;
+  guint current_cache_serial;
+  /* A cache of size four should be more than enough, people usually
+   * have two groups around, and the xkb limit is four.  It still
+   * works correct for more than four groups.  It's just the
+   * cache.
+   */
+  DirectionCacheEntry group_direction_cache[4];
 #endif
 };
 
 #define KEYMAP_USE_XKB(keymap) GDK_DISPLAY_X11 ((keymap)->display)->use_xkb
 #define KEYMAP_XDISPLAY(keymap) GDK_DISPLAY_XDISPLAY ((keymap)->display)
 
-static GType gdk_keymap_x11_get_type (void);
-static void  gdk_keymap_x11_init     (GdkKeymapX11 *keymap);
+static GType gdk_keymap_x11_get_type   (void);
+static void  gdk_keymap_x11_class_init (GdkKeymapX11Class *klass);
+static void  gdk_keymap_x11_init       (GdkKeymapX11      *keymap);
+static void  gdk_keymap_x11_finalize   (GObject           *object);
+
+static GdkKeymapClass *parent_class = NULL;
 
 static GType
 gdk_keymap_x11_get_type (void)
@@ -98,7 +123,7 @@ gdk_keymap_x11_get_type (void)
 	  sizeof (GdkKeymapClass),
 	  (GBaseInitFunc) NULL,
 	  (GBaseFinalizeFunc) NULL,
-	  (GClassInitFunc) NULL,
+	  (GClassInitFunc) gdk_keymap_x11_class_init,
 	  NULL,           /* class_finalize */
 	  NULL,           /* class_data */
 	  sizeof (GdkKeymapX11),
@@ -107,13 +132,23 @@ gdk_keymap_x11_get_type (void)
 	};
       
       object_type = g_type_register_static (GDK_TYPE_KEYMAP,
-                                            "GdkKeymapX11",
+                                            g_intern_static_string ("GdkKeymapX11"),
                                             &object_info, 0);
     }
   
   return object_type;
 }
 
+static void
+gdk_keymap_x11_class_init (GdkKeymapX11Class *klass)
+{
+  GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
+  parent_class = g_type_class_peek_parent (klass);
+
+  object_class->finalize = gdk_keymap_x11_finalize;
+}
+
 static void
 gdk_keymap_x11_init (GdkKeymapX11 *keymap)
 {
@@ -129,12 +164,33 @@ gdk_keymap_x11_init (GdkKeymapX11 *keymap)
   keymap->group_switch_mask = 0;
   keymap->lock_keysym = GDK_Caps_Lock;
   keymap->have_direction = FALSE;
-  
+  keymap->current_serial = 0;
+
 #ifdef HAVE_XKB
   keymap->xkb_desc = NULL;
+  keymap->current_group_atom = 0;
+  keymap->current_cache_serial = 0;
 #endif
 
-  keymap->current_serial = 0;
+}
+
+static void
+gdk_keymap_x11_finalize (GObject *object)
+{
+  GdkKeymapX11 *keymap_x11 = GDK_KEYMAP_X11 (object);
+
+  if (keymap_x11->keymap)
+    XFree (keymap_x11->keymap);
+
+  if (keymap_x11->mod_keymap)
+    XFreeModifiermap (keymap_x11->mod_keymap);
+
+#ifdef HAVE_XKB
+  if (keymap_x11->xkb_desc)
+    XkbFreeKeyboard (keymap_x11->xkb_desc, XkbAllComponentsMask, True);
+#endif
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static inline void
@@ -147,6 +203,46 @@ update_keyrange (GdkKeymapX11 *keymap_x11)
 
 #ifdef HAVE_XKB
 
+static void
+update_modmap (Display      *display,
+	       GdkKeymapX11 *keymap_x11)
+{
+  static struct {
+    const gchar *name;
+    Atom atom;
+    GdkModifierType mask;
+  } vmods[] = {
+    { "Meta", 0, GDK_META_MASK },
+    { "Super", 0, GDK_SUPER_MASK },
+    { "Hyper", 0, GDK_HYPER_MASK },
+    { NULL, 0, 0 }
+  };
+
+  gint i, j, k;
+
+  if (!vmods[0].atom)
+    for (i = 0; vmods[i].name; i++)
+      vmods[i].atom = XInternAtom (display, vmods[i].name, FALSE);
+
+  for (i = 0; i < 8; i++)
+    keymap_x11->modmap[i] = 1 << i;
+
+  for (i = 0; i < XkbNumVirtualMods; i++)
+    {
+      for (j = 0; vmods[j].atom; j++)
+	{
+	  if (keymap_x11->xkb_desc->names->vmods[i] == vmods[j].atom)
+	    {
+	      for (k = 0; k < 8; k++)
+		{
+		  if (keymap_x11->xkb_desc->server->vmods[i] & (1 << k))
+		    keymap_x11->modmap[k] |= vmods[j].mask;
+		}
+	    }
+	}
+    }
+}
+
 static XkbDescPtr
 get_xkb (GdkKeymapX11 *keymap_x11)
 {
@@ -157,17 +253,24 @@ get_xkb (GdkKeymapX11 *keymap_x11)
   
   if (keymap_x11->xkb_desc == NULL)
     {
-      keymap_x11->xkb_desc = XkbGetMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask, XkbUseCoreKbd);
+      keymap_x11->xkb_desc = XkbGetMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask | XkbModifierMapMask | XkbVirtualModsMask, XkbUseCoreKbd);
       if (keymap_x11->xkb_desc == NULL)
-	g_error ("Failed to get keymap");
+        {
+	  g_error ("Failed to get keymap");
+          return NULL;
+        }
 
-      XkbGetNames (xdisplay, XkbGroupNamesMask, keymap_x11->xkb_desc);
+      XkbGetNames (xdisplay, XkbGroupNamesMask | XkbVirtualModNamesMask, keymap_x11->xkb_desc);
+
+      update_modmap (xdisplay, keymap_x11);
     }
   else if (keymap_x11->current_serial != display_x11->keymap_serial)
     {
-      XkbGetUpdatedMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask,
+      XkbGetUpdatedMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask | XkbModifierMapMask | XkbVirtualModsMask,
 			keymap_x11->xkb_desc);
-      XkbGetNames (xdisplay, XkbGroupNamesMask, keymap_x11->xkb_desc);
+      XkbGetNames (xdisplay, XkbGroupNamesMask | XkbVirtualModNamesMask, keymap_x11->xkb_desc);
+
+      update_modmap (xdisplay, keymap_x11);
     }
 
   keymap_x11->current_serial = display_x11->keymap_serial;
@@ -210,22 +313,41 @@ gdk_keymap_get_for_display (GdkDisplay *display)
  * otherwise we lose a whole group of keys
  */
 #define KEYSYM_INDEX(keymap_impl, group, level) \
-  (2 * ((group) % (int)((keymap_impl->keysyms_per_keycode + 1) / 2)) + (level))
+  (2 * ((group) % (gint)((keymap_impl->keysyms_per_keycode + 1) / 2)) + (level))
 #define KEYSYM_IS_KEYPAD(s) (((s) >= 0xff80 && (s) <= 0xffbd) || \
                              ((s) >= 0x11000000 && (s) <= 0x1100ffff))
 
-static int
-get_symbol (const KeySym *syms, GdkKeymapX11 *keymap_x11, int group, int level)
+static gint
+get_symbol (const KeySym *syms,
+	    GdkKeymapX11 *keymap_x11,
+	    gint group,
+	    gint level)
 {
-  int index;
+  gint index;
 
   index = KEYSYM_INDEX(keymap_x11, group, level);
-  if (index > keymap_x11->keysyms_per_keycode)
+  if (index >= keymap_x11->keysyms_per_keycode)
       return NoSymbol;
 
   return syms[index];
 }
 
+static void
+set_symbol (KeySym       *syms, 
+	    GdkKeymapX11 *keymap_x11, 
+	    gint          group, 
+	    gint          level, 
+	    KeySym        sym)
+{
+  gint index;
+
+  index = KEYSYM_INDEX(keymap_x11, group, level);
+  if (index >= keymap_x11->keysyms_per_keycode)
+      return;
+
+  syms[index] = sym;
+}
+
 static void
 update_keymaps (GdkKeymapX11 *keymap_x11)
 {
@@ -271,7 +393,7 @@ update_keymaps (GdkKeymapX11 *keymap_x11)
 	  for (i = 0 ; i < 2 ; i++)
 	    {
 	      if (get_symbol (syms, keymap_x11, i, 0) == GDK_Tab)
-		syms[KEYSYM_INDEX (keymap_x11, i, 1)] = GDK_ISO_Left_Tab;
+		set_symbol (syms, keymap_x11, i, 1, GDK_ISO_Left_Tab);
 	    }
 
           /*
@@ -286,8 +408,8 @@ update_keymaps (GdkKeymapX11 *keymap_x11)
               gdk_keyval_convert_case (get_symbol (syms, keymap_x11, 0, 0), &lower, &upper);
               if (lower != upper)
                 {
-                  syms[KEYSYM_INDEX (keymap_x11, 0, 0)] = lower;
-                  syms[KEYSYM_INDEX (keymap_x11, 0, 1)] = upper;
+		  set_symbol (syms, keymap_x11, 0, 0, lower);
+		  set_symbol (syms, keymap_x11, 0, 1, upper);
                 }
             }
       
@@ -301,6 +423,9 @@ update_keymaps (GdkKeymapX11 *keymap_x11)
       keymap_x11->group_switch_mask = 0;
       keymap_x11->num_lock_mask = 0;
 
+      for (i = 0; i < 8; i++)
+	keymap_x11->modmap[i] = 1 << i;
+      
       /* There are 8 sets of modifiers, with each set containing
        * max_keypermod keycodes.
        */
@@ -320,6 +445,22 @@ update_keymaps (GdkKeymapX11 *keymap_x11)
 
           syms = keymap_x11->keymap + (keycode - keymap_x11->min_keycode) * keymap_x11->keysyms_per_keycode;
 
+	  mask = 0;
+	  for (j = 0; j < keymap_x11->keysyms_per_keycode; j++)
+	    {
+	      if (syms[j] == GDK_Meta_L ||
+		  syms[j] == GDK_Meta_R)
+		mask |= GDK_META_MASK;
+	      else if (syms[j] == GDK_Hyper_L ||
+		       syms[j] == GDK_Hyper_R)
+		mask |= GDK_HYPER_MASK;
+	      else if (syms[j] == GDK_Super_L ||
+		       syms[j] == GDK_Super_R)
+		mask |= GDK_SUPER_MASK;
+	    }
+
+	  keymap_x11->modmap[i/keymap_x11->mod_keymap->max_keypermod] |= mask;
+
           /* The fourth modifier, GDK_MOD1_MASK is 1 << 3.
 	   * Each group of max_keypermod entries refers to the same modifier.
            */
@@ -396,7 +537,7 @@ get_keymap (GdkKeymapX11 *keymap_x11)
 
 static GdkKeymap *
 get_effective_keymap (GdkKeymap  *keymap,
-		       const char *function)
+		      const char *function)
 {
   if (!keymap)
     {
@@ -411,51 +552,135 @@ get_effective_keymap (GdkKeymap  *keymap,
 
 #if HAVE_XKB
 static PangoDirection
-get_direction (GdkKeymapX11 *keymap_x11)
+get_direction (XkbDescRec *xkb, 
+	       gint group)
 {
-  XkbDescRec *xkb = get_xkb (keymap_x11);
-  const char *name;
-  XkbStateRec state_rec;
-  PangoDirection result;
+  gint code;
 
-  GdkDisplay *display = GDK_KEYMAP (keymap_x11)->display;
+  gint rtl_minus_ltr = 0; /* total number of RTL keysyms minus LTR ones */
 
-  XkbGetState (GDK_DISPLAY_XDISPLAY (display), XkbUseCoreKbd, &state_rec);
-  
-  if (xkb->names->groups[state_rec.locked_group] == None)
-    result = PANGO_DIRECTION_LTR;
-  else
+  for (code = xkb->min_key_code; code <= xkb->max_key_code; code++)
     {
-      name = gdk_x11_get_xatom_name_for_display (display, xkb->names->groups[state_rec.locked_group]);
+      gint width = XkbKeyGroupWidth (xkb, code, group);
+      gint level;
+      for (level = 0; level < width; level++)
+	{
+	  KeySym sym = XkbKeySymEntry (xkb, code, level, group);
+	  PangoDirection dir = pango_unichar_direction (gdk_keyval_to_unicode (sym));
 
-      if (g_ascii_strcasecmp (name, "arabic") == 0 ||
-	  g_ascii_strcasecmp (name, "hebrew") == 0 ||
-	  g_ascii_strcasecmp (name, "israelian") == 0)
-	result = PANGO_DIRECTION_RTL;
-      else
-	result = PANGO_DIRECTION_LTR;
+	  switch (dir)
+	    {
+	    case PANGO_DIRECTION_RTL:
+	      rtl_minus_ltr++;
+	      break;
+	    case PANGO_DIRECTION_LTR:
+	      rtl_minus_ltr--;
+	      break;
+	    default:
+	      break;
+	    }
+	}
     }
     
-  return result;
+  if (rtl_minus_ltr > 0)
+    return PANGO_DIRECTION_RTL;
+  else
+    return PANGO_DIRECTION_LTR;
 }
 
+static void
+update_direction (GdkKeymapX11 *keymap_x11,
+		  gint          group)
+{
+  XkbDescRec *xkb = get_xkb (keymap_x11);
+  Atom group_atom;
+
+  group_atom = xkb->names->groups[group];
+
+  /* a group change? */
+  if (!keymap_x11->have_direction || keymap_x11->current_group_atom != group_atom)
+    {
+      gboolean cache_hit = FALSE;
+      DirectionCacheEntry *cache = keymap_x11->group_direction_cache;
+
+      PangoDirection direction = PANGO_DIRECTION_NEUTRAL;
+      gint i;
+
+      if (keymap_x11->have_direction)
+	{
+          /* lookup in cache */
+	  for (i = 0; i < G_N_ELEMENTS (keymap_x11->group_direction_cache); i++)
+	  {
+	    if (cache[i].group_atom == group_atom)
+	      {
+		cache_hit = TRUE;
+		cache[i].serial = keymap_x11->current_cache_serial++; /* freshen */
+		direction = cache[i].direction;
+		group_atom = cache[i].group_atom;
+		break;
+	      }
+	  }
+	}
+      else
+	{
+          /* initialize cache */
+	  for (i = 0; i < G_N_ELEMENTS (keymap_x11->group_direction_cache); i++)
+	    {
+	      cache[i].group_atom = 0;
+	      cache[i].serial = keymap_x11->current_cache_serial;
+	    }
+	  keymap_x11->current_cache_serial++;
+	}
+
+      /* insert in cache */
+      if (!cache_hit)
+	{
+	  gint oldest = 0;
+
+	  direction = get_direction (xkb, group);
+
+	  /* remove the oldest entry */
+	  for (i = 0; i < G_N_ELEMENTS (keymap_x11->group_direction_cache); i++)
+	    {
+	      if (cache[i].serial < cache[oldest].serial)
+	        oldest = i;
+	    }
+	  
+	  cache[oldest].group_atom = group_atom;
+	  cache[oldest].direction = direction;
+	  cache[oldest].serial = keymap_x11->current_cache_serial++;
+	}
+
+      keymap_x11->current_group_atom = group_atom;
+
+      keymap_x11->have_direction = TRUE;
+      keymap_x11->current_direction = direction;
+    }
+}
+
+/* keep this in sync with the XkbSelectEventDetails() call 
+ * in gdk_display_open()
+ */
 void
-_gdk_keymap_state_changed (GdkDisplay *display)
+_gdk_keymap_state_changed (GdkDisplay *display,
+			   XEvent     *xevent)
 {
   GdkDisplayX11 *display_x11 = GDK_DISPLAY_X11 (display);
+  XkbEvent *xkb_event = (XkbEvent *)xevent;
   
   if (display_x11->keymap)
     {
       GdkKeymapX11 *keymap_x11 = GDK_KEYMAP_X11 (display_x11->keymap);
+      gboolean had_direction;
+      PangoDirection direction;
       
-      PangoDirection new_direction = get_direction (keymap_x11);
+      had_direction = keymap_x11->have_direction;
+      direction = keymap_x11->current_direction;
       
-      if (!keymap_x11->have_direction || new_direction != keymap_x11->current_direction)
-	{
-	  keymap_x11->have_direction = TRUE;
-	  keymap_x11->current_direction = new_direction;
-	  g_signal_emit_by_name (keymap_x11, "direction_changed");
-	}
+      update_direction (keymap_x11, xkb_event->state.locked_group);
+      
+      if (!had_direction || direction != keymap_x11->current_direction)
+	g_signal_emit_by_name (keymap_x11, "direction_changed");      
     }
 }
 
@@ -481,18 +706,22 @@ gdk_keymap_get_direction (GdkKeymap *keymap)
   if (KEYMAP_USE_XKB (keymap))
     {
       GdkKeymapX11 *keymap_x11 = GDK_KEYMAP_X11 (keymap);
-      
+
       if (!keymap_x11->have_direction)
 	{
-	  keymap_x11->current_direction = get_direction (keymap_x11);
-	  keymap_x11->have_direction = TRUE;
+	  GdkDisplay *display = GDK_KEYMAP (keymap_x11)->display;
+	  XkbStateRec state_rec;
+
+	  XkbGetState (GDK_DISPLAY_XDISPLAY (display), XkbUseCoreKbd, 
+		       &state_rec);
+	  update_direction (keymap_x11, XkbGroupLock (&state_rec));
 	}
   
       return keymap_x11->current_direction;
     }
   else
 #endif /* HAVE_XKB */
-    return PANGO_DIRECTION_LTR;
+    return PANGO_DIRECTION_NEUTRAL;
 }
 
 /**
@@ -786,36 +1015,20 @@ gdk_keymap_get_entries_for_keycode (GdkKeymap     *keymap,
           ++i;
         }
     }
+
+  *n_entries = 0;
+
+  if (keys)
+    {
+      *n_entries = key_array->len;
+      *keys = (GdkKeymapKey*) g_array_free (key_array, FALSE);
+    }
   
-  if ((key_array && key_array->len > 0) ||
-      (keyval_array && keyval_array->len > 0))
+  if (keyvals)
     {
-      if (keys)
-        *keys = (GdkKeymapKey*) key_array->data;
-
-      if (keyvals)
-        *keyvals = (guint*) keyval_array->data;
-
-      if (key_array)
-        *n_entries = key_array->len;
-      else
-        *n_entries = keyval_array->len;
+      *n_entries = keyval_array->len;
+      *keyvals = (guint*) g_array_free (keyval_array, FALSE);
     }
-  else
-    {
-      if (keys)
-        *keys = NULL;
-
-      if (keyvals)
-        *keyvals = NULL;
-      
-      *n_entries = 0;
-    }
-
-  if (key_array)
-    g_array_free (key_array, key_array->len > 0 ? FALSE : TRUE);
-  if (keyval_array)
-    g_array_free (keyval_array, keyval_array->len > 0 ? FALSE : TRUE);
 
   return *n_entries > 0;
 }
@@ -991,7 +1204,7 @@ MyEnhancedXkbTranslateKeyCode(register XkbDescPtr     xkb,
     
     /* ---- End stuff GDK adds to the original Xlib version ---- */
     
-    return (syms[col]!=NoSymbol);
+    return (syms[col] != NoSymbol);
 }
 #endif /* HAVE_XKB */
 
@@ -1058,7 +1271,7 @@ translate_keysym (GdkKeymapX11   *keymap_x11,
     }
   else
     {
-      /* Fall back to the the first level if no symbol for the level
+      /* Fall back to the first level if no symbol for the level
        * we were passed.
        */
       shift_level = (state & shift_modifiers) ? 1 : 0;
@@ -1187,7 +1400,7 @@ gdk_keymap_translate_keyboard_state (GdkKeymap       *keymap,
   if (hardware_keycode < keymap_x11->min_keycode ||
       hardware_keycode > keymap_x11->max_keycode)
     return FALSE;
-  
+
 #ifdef HAVE_XKB
   if (KEYMAP_USE_XKB (keymap))
     {
@@ -1334,3 +1547,68 @@ _gdk_x11_get_group_for_state (GdkDisplay      *display,
       return (state & keymap_impl->group_switch_mask) ? 1 : 0;
     }
 }
+
+void
+_gdk_keymap_add_virtual_modifiers (GdkKeymap       *keymap,
+				   GdkModifierType *modifiers)
+{
+  GdkKeymapX11 *keymap_x11;
+  int i;
+  
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);
+  keymap_x11 = GDK_KEYMAP_X11 (keymap);
+
+  for (i = 3; i < 8; i++)
+    {
+      if ((1 << i) & *modifiers)
+        {
+	  if (keymap_x11->modmap[i] & GDK_MOD1_MASK)
+	    *modifiers |= GDK_MOD1_MASK;
+	  else if (keymap_x11->modmap[i] & GDK_SUPER_MASK)
+	    *modifiers |= GDK_SUPER_MASK;
+	  else if (keymap_x11->modmap[i] & GDK_HYPER_MASK)
+	    *modifiers |= GDK_HYPER_MASK;
+	  else if (keymap_x11->modmap[i] & GDK_META_MASK)
+	    *modifiers |= GDK_META_MASK;
+        }
+    }
+}
+
+gboolean
+_gdk_keymap_key_is_modifier (GdkKeymap *keymap,
+			     guint      keycode)
+{
+  GdkKeymapX11 *keymap_x11;
+  gint i;
+
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);  
+  keymap_x11 = GDK_KEYMAP_X11 (keymap);
+
+  if (keycode < keymap_x11->min_keycode ||
+      keycode > keymap_x11->max_keycode)
+    return FALSE;
+
+#ifdef HAVE_XKB
+  if (KEYMAP_USE_XKB (keymap))
+    {
+      XkbDescRec *xkb = get_xkb (keymap_x11);
+      
+      if (xkb->map->modmap[keycode] != 0)
+	return TRUE;
+    }
+  else
+#endif
+    {
+      for (i = 0; i < 8 * keymap_x11->mod_keymap->max_keypermod; i++)
+	{
+	  if (keycode == keymap_x11->mod_keymap->modifiermap[i])
+	    return TRUE;
+	}
+    }
+
+  return FALSE;
+}
+
+
+#define __GDK_KEYS_X11_C__
+#include "gdkaliasdef.c"
