@@ -31,6 +31,8 @@
 #include <config.h>
 
 #include "gtkactiongroup.h"
+#include "gtkiconfactory.h"
+#include "gtkicontheme.h"
 #include "gtkstock.h"
 #include "gtktoggleaction.h"
 #include "gtkradioaction.h"
@@ -106,7 +108,7 @@ gtk_action_group_get_type (void)
         (GInstanceInitFunc) gtk_action_group_init,
       };
 
-      type = g_type_register_static (G_TYPE_OBJECT, "GtkActionGroup",
+      type = g_type_register_static (G_TYPE_OBJECT, I_("GtkActionGroup"),
 				     &type_info, 0);
     }
 
@@ -172,7 +174,7 @@ gtk_action_group_class_init (GtkActionGroupClass *klass)
    * Since: 2.4
    */
   action_group_signals[CONNECT_PROXY] =
-    g_signal_new ("connect_proxy",
+    g_signal_new (I_("connect_proxy"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  0, 0, NULL, NULL,
 		  _gtk_marshal_VOID__OBJECT_OBJECT,
@@ -195,7 +197,7 @@ gtk_action_group_class_init (GtkActionGroupClass *klass)
    * Since: 2.4
    */
   action_group_signals[DISCONNECT_PROXY] =
-    g_signal_new ("disconnect_proxy",
+    g_signal_new (I_("disconnect_proxy"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  0, 0, NULL, NULL,
 		  _gtk_marshal_VOID__OBJECT_OBJECT,
@@ -216,7 +218,7 @@ gtk_action_group_class_init (GtkActionGroupClass *klass)
    * Since: 2.4
    */
   action_group_signals[PRE_ACTIVATE] =
-    g_signal_new ("pre_activate",
+    g_signal_new (I_("pre_activate"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  0, 0, NULL, NULL,
 		  _gtk_marshal_VOID__OBJECT,
@@ -237,7 +239,7 @@ gtk_action_group_class_init (GtkActionGroupClass *klass)
    * Since: 2.4
    */
   action_group_signals[POST_ACTIVATE] =
-    g_signal_new ("post_activate",
+    g_signal_new (I_("post_activate"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  0, 0, NULL, NULL,
 		  _gtk_marshal_VOID__OBJECT,
@@ -251,7 +253,7 @@ gtk_action_group_class_init (GtkActionGroupClass *klass)
 static void 
 remove_action (GtkAction *action) 
 {
-  g_object_set (action, "action-group", NULL, NULL);
+  g_object_set (action, I_("action-group"), NULL, NULL);
   g_object_unref (action);
 }
 
@@ -420,10 +422,10 @@ static void
 cb_set_action_sensitivity (const gchar *name, 
 			   GtkAction   *action)
 {
-  /* Minor optimization, the action_groups state only affects actions that are
-   * themselves sensitive */
+  /* Minor optimization, the action_groups state only affects actions 
+   * that are themselves sensitive */
   if (gtk_action_get_sensitive (action))
-    g_object_notify (G_OBJECT (action), "sensitive");
+    _gtk_action_sync_sensitive (action);
 }
 
 /**
@@ -478,10 +480,10 @@ static void
 cb_set_action_visiblity (const gchar *name, 
 			 GtkAction   *action)
 {
-  /* Minor optimization, the action_groups state only affects actions that are
-   * themselves sensitive */
+  /* Minor optimization, the action_groups state only affects actions 
+   * that are themselves visible */
   if (gtk_action_get_visible (action))
-    g_object_notify (G_OBJECT (action), "visible");
+    _gtk_action_sync_visible (action);
 }
 
 /**
@@ -558,7 +560,7 @@ gtk_action_group_add_action (GtkActionGroup *action_group,
   g_hash_table_insert (action_group->private_data->actions, 
 		       g_strdup (gtk_action_get_name (action)),
                        g_object_ref (action));
-  g_object_set (action, "action-group", action_group, NULL);
+  g_object_set (action, I_("action-group"), action_group, NULL);
 }
 
 /**
@@ -726,7 +728,7 @@ shared_data_unref (gpointer data)
       if (shared_data->destroy) 
 	(*shared_data->destroy) (shared_data->data);
       
-      g_free (shared_data);
+      g_slice_free (SharedData, shared_data);
     }
 }
 
@@ -760,7 +762,7 @@ gtk_action_group_add_actions_full (GtkActionGroup       *action_group,
 
   g_return_if_fail (GTK_IS_ACTION_GROUP (action_group));
 
-  shared_data = g_new0 (SharedData, 1);
+  shared_data = g_slice_new0 (SharedData);
   shared_data->ref_count = 1;
   shared_data->data = user_data;
   shared_data->destroy = destroy;
@@ -777,8 +779,16 @@ gtk_action_group_add_actions_full (GtkActionGroup       *action_group,
       action = gtk_action_new (entries[i].name,
 			       label,
 			       tooltip,
-			       entries[i].stock_id);
+			       NULL);
 
+      if (entries[i].stock_id) 
+	{
+	  g_object_set (action, "stock-id", entries[i].stock_id, NULL);
+	  if (gtk_icon_theme_has_icon (gtk_icon_theme_get_default (), 
+				       entries[i].stock_id))
+	    g_object_set (action, "icon-name", entries[i].stock_id, NULL);
+	}
+	  
       if (entries[i].callback)
 	{
 	  GClosure *closure;
@@ -856,7 +866,7 @@ gtk_action_group_add_toggle_actions_full (GtkActionGroup             *action_gro
 
   g_return_if_fail (GTK_IS_ACTION_GROUP (action_group));
 
-  shared_data = g_new0 (SharedData, 1);
+  shared_data = g_slice_new0 (SharedData);
   shared_data->ref_count = 1;
   shared_data->data = user_data;
   shared_data->destroy = destroy;
@@ -873,7 +883,15 @@ gtk_action_group_add_toggle_actions_full (GtkActionGroup             *action_gro
       action = gtk_toggle_action_new (entries[i].name,
 				      label,
 				      tooltip,
-				      entries[i].stock_id);
+				      NULL);
+
+      if (entries[i].stock_id) 
+	{
+	  if (gtk_icon_factory_lookup_default (entries[i].stock_id))
+	    g_object_set (action, "stock-id", entries[i].stock_id, NULL);
+	  else
+	    g_object_set (action, "icon-name", entries[i].stock_id, NULL);
+	}
 
       gtk_toggle_action_set_active (action, entries[i].is_active);
 
@@ -977,9 +995,17 @@ gtk_action_group_add_radio_actions_full (GtkActionGroup            *action_group
       action = gtk_radio_action_new (entries[i].name,
 				     label,
 				     tooltip,
-				     entries[i].stock_id,
+				     NULL,
 				     entries[i].value);
 
+      if (entries[i].stock_id) 
+	{
+	  if (gtk_icon_factory_lookup_default (entries[i].stock_id))
+	    g_object_set (action, "stock-id", entries[i].stock_id, NULL);
+	  else
+	    g_object_set (action, "icon-name", entries[i].stock_id, NULL);
+	}
+
       if (i == 0) 
 	first_action = action;
 
@@ -1038,8 +1064,8 @@ dgettext_swapped (const gchar *msgid,
 		  const gchar *domainname)
 {
   /* Pass through dgettext if and only if msgid is nonempty. */
-  if (msgid && *msgid)
-    return dgettext (domainname, msgid);
+  if (msgid && *msgid) 
+    return dgettext (domainname, msgid); 
   else
     return (gchar*) msgid;
 }
